/*==============================================================================*/
/*  Copyright (C) 2020, Heartland Data inc. All Rights Reserved.                */
/*                                                                              */
/*  Title  :   i2c Driver                                                       */
/*  EventID:   Insert Kernelinfo Test Point                                     */
/*  AppVer :   DT+ Ver1.00`                                                    */
/*  FileID :   3bb                                                              */
/*  Version:   1.0                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Please customize the code for your environment.                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for PORT Control                                             */
/*==============================================================================*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/version.h>

#define LICENSE "GPL v2"

/*==============================================================================*/
/*  Macro:  DT_UINT                                                             */
/*  Desc:   Please change Test Point argument type for DT+ Project setting.     */
/*==============================================================================*/
#define	DT_UINT unsigned int

/*==============================================================================*/
/*  Macro:  PROC_NAME                                                           */
/*  Desc:   Register Proc File name for Linux.                                  */
/*==============================================================================*/
#define	PROC_NAME "dtproc"

/*==============================================================================*/
/*  Struct:  DT_DATA                                                            */
/*  Desc:    Test Point Data Struct                                             */
/*==============================================================================*/
struct DT_DATA {
	char tp_func;
	DT_UINT addr;
	DT_UINT dat;
	DT_UINT size;
	unsigned char value[256];
};

/*==============================================================================*/
/*  Desc:    DT_DATA.tp_func ID                                                 */
/*==============================================================================*/
#define TP_BusOut		'2'
#define TP_MemoryOutput		'3'
#define TP_BusOutFast		'4'
#define TP_MemoryOutputFast		'5'
#define TP_EventTrigger		'6'
#define TP_EventTriggerFast		'7'
#define TP_EventTrigger32		'8'
#define TP_EventTrigger32Fast		'9'

/*==============================================================================*/
/*  Desc:   Test Point Parametaer                                               */
/*==============================================================================*/
#define		DT_VARIABLE_BIT			0x02
#define		DT_EVTTRG_BIT			0x08
#define		DT_VARIABLE_FAST_BIT		0x01
#define		DT_EVTTRG_FAST_BIT		0x02

/*==============================================================================*/
/*  Macro:  DT_USE_SPINLOCK                                                     */
/*  Desc:   Please set 1, when use spin_lock_irqsave function.                  */
/*==============================================================================*/
#include <linux/spinlock.h>
static DEFINE_SPINLOCK(tplock);
unsigned long flags = 0;
#define DT_LOCK	spin_lock_irqsave(&tplock, flags)
#define DT_UNLOCK	spin_unlock_irqrestore(&tplock, flags)

/*==============================================================================*/
/*  Desc:   GPIO Control setting                                                */
/*==============================================================================*/
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000)

#define	CLK_BIT		25	/* CLK */
#define	DAT0_BIT	8	/* DAT0 */

void *iop = NULL;

#define	INP_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) & ~(7<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define OUT_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) | (1<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define	GPIO_SET(g) iowrite32( 1<<(g), (void*)((int *)iop+7) )
#define	GPIO_CLR(g) iowrite32( 1<<(g), (void*)((int *)iop+10) )
#define	GPIO_GET ioread32((void*)((int *)iop+13))

#define	START_CON	{			\
	GPIO_CLR(DAT0_BIT);			\
}
#define	STOP_CON	{			\
	GPIO_SET(DAT0_BIT);			\
}
#define RESTART_CON	{			\
	GPIO_CLR(CLK_BIT);			\
	GPIO_SET(DAT0_BIT);			\
	GPIO_SET(CLK_BIT);			\
	GPIO_CLR(DAT0_BIT);			\
}

#define	DATA_OUT(x)	{			\
	GPIO_CLR(CLK_BIT);			\
	if( (x) ){				\
		GPIO_SET(DAT0_BIT);		\
	}					\
	else{					\
		GPIO_CLR(DAT0_BIT);		\
	}					\
	GPIO_SET(CLK_BIT);			\
}

static int init = 0;	/* init flag */
static unsigned int start_trace = 0;
/*==============================================================================*/
/*  Func:   _TP_BusPortInit                                                     */
/*  Desc:   Please describe the code to initializes ports.                      */
/*==============================================================================*/
static void _TP_BusPortInit(void)
{
	unsigned int reg;

	if (init == 0) {
		iop = ioremap(GPIO_BASE,16);
		if (iop == NULL) {
			printk("dtdrvspi: init err\n");
			return;
		}
		OUT_GPIO(CLK_BIT);
		OUT_GPIO(DAT0_BIT);

		GPIO_SET(CLK_BIT);
		init = 1;
		return;
	}
}

/*==============================================================================*/
/*  Func:   _TP_BusOutDrv                                                       */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static inline void _TP_BusOutDrv( DT_UINT addr, DT_UINT data )
{
	DATA_OUT(data & 0x8000);	/* bit15 */
	DATA_OUT(data & 0x4000);	/* bit14 */
	DATA_OUT(data & 0x2000);	/* bit13 */
	DATA_OUT(data & 0x1000);	/* bit12 */
	DATA_OUT(data & 0x0800);	/* bit11 */
	DATA_OUT(data & 0x0400);	/* bit10 */
	DATA_OUT(data & 0x0200);	/* bit9 */
	DATA_OUT(data & 0x0100);	/* bit8 */
	DATA_OUT(0);				/* ACK */
	DATA_OUT(data & 0x0080);	/* bit7 */
	DATA_OUT(data & 0x0040);	/* bit6 */
	DATA_OUT(data & 0x0020);	/* bit5 */
	DATA_OUT(data & 0x0010);	/* bit4 */
	DATA_OUT(data & 0x0008);	/* bit3 */
	DATA_OUT(data & 0x0004);	/* bit2 */
	DATA_OUT(data & 0x0002);	/* bit1 */
	DATA_OUT(data & 0x0001);	/* bit0 */
	DATA_OUT(0);				/* ACK */

	DATA_OUT(addr & 0x8000);	/* bit15 */
	DATA_OUT(addr & 0x4000);	/* bit14 */
	DATA_OUT(addr & 0x2000);	/* bit13 */
	DATA_OUT(addr & 0x1000);	/* bit12 */
	DATA_OUT(addr & 0x0800);	/* bit11 */
	DATA_OUT(addr & 0x0400);	/* bit10 */
	DATA_OUT(addr & 0x0200);	/* bit9 */
	DATA_OUT(addr & 0x0100);	/* bit8 */
	DATA_OUT(0);				/* ACK */
	DATA_OUT(addr & 0x0080);	/* bit7 */
	DATA_OUT(addr & 0x0040);	/* bit6 */
	DATA_OUT(addr & 0x0020);	/* bit5 */
	DATA_OUT(addr & 0x0010);	/* bit4 */
	DATA_OUT(addr & 0x0008);	/* bit3 */
	DATA_OUT(addr & 0x0004);	/* bit2 */
	DATA_OUT(addr & 0x0002);	/* bit1 */
	DATA_OUT(addr & 0x0001);	/* bit0 */
	DATA_OUT(0);				/* ACK */
}

/*==============================================================================*/
/*  Func:   _TP_MemoryOutDrv                                                    */
/*  Desc:   Value Output Function.                                              */
/*==============================================================================*/
static inline void _TP_MemoryOutDrv(unsigned char data)
{
	DATA_OUT(data & 0x0080);	/* bit7 */
	DATA_OUT(data & 0x0040);	/* bit6 */
	DATA_OUT(data & 0x0020);	/* bit5 */
	DATA_OUT(data & 0x0010);	/* bit4 */
	DATA_OUT(data & 0x0008);	/* bit3 */
	DATA_OUT(data & 0x0004);	/* bit2 */
	DATA_OUT(data & 0x0002);	/* bit1 */
	DATA_OUT(data & 0x0001);	/* bit0 */
	DATA_OUT(0);				/* ACK */
}

/*==============================================================================*/
/*  Func:   _TP_BusOutFastDrv                                                   */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static void _TP_BusOutFastDrv(DT_UINT data, DT_UINT bit_num)
{
	if (bit_num >= 8) {
		_TP_MemoryOutDrv(data >> 8  );
	}
	_TP_MemoryOutDrv(data);
}

/**********************************kernel TestPoint API**********************************/

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut( DT_UINT addr, DT_UINT dat )
{
	if (start_trace != 1) {
		return;
	}
	DT_LOCK;
	START_CON;
	_TP_BusOutDrv(addr, dat);
	STOP_CON;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	if (size > 256) size = 256;
	DT_LOCK;
	START_CON;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT), dat);
	RESTART_CON;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	STOP_CON;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_BusOutFast                                                      */
/*  Desc:   Called by Fast Test Point                                           */
/*==============================================================================*/
void _TP_BusOutFast( DT_UINT step, DT_UINT bit_num )
{
	if (start_trace != 1) {
		return;
	}

	step = step << 2;
	bit_num = bit_num + 2;
	DT_LOCK;
	START_CON;
	_TP_BusOutFastDrv(step, bit_num);
	STOP_CON;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusOutFast);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutputFast                                                */
/*  Desc:   Called by Variable Fast Test Point                                  */
/*==============================================================================*/
void _TP_MemoryOutputFast( DT_UINT step, DT_UINT bit_num, void *value, DT_UINT size)
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	step = (step << 2) + DT_VARIABLE_FAST_BIT;
	bit_num = bit_num + 2;
	if (size > 256) size = 256;
	DT_LOCK;
	START_CON;
	_TP_BusOutFastDrv(step, bit_num);
	RESTART_CON;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	STOP_CON;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutputFast);

/*==============================================================================*/
/*  Func:   _TP_BusKernelInfo                                                   */
/*  Desc:   Called by KernelInfo Test Point.                                    */
/*==============================================================================*/
void _TP_BusKernelInfo(DT_UINT addr, DT_UINT dat, struct task_struct *next)
{
	int i;
	unsigned int pid = next->pid;
	unsigned int tgid = next->mm?next->tgid:0;
	unsigned char *p = next->comm;

	if (start_trace != 1) {
		return;
	}

	DT_LOCK;
	START_CON;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT | DT_EVTTRG_BIT), dat);
	RESTART_CON;
	_TP_MemoryOutDrv(pid);
	_TP_MemoryOutDrv((pid >> 8));
	_TP_MemoryOutDrv((pid >> 16));
	_TP_MemoryOutDrv((pid >> 24));
	_TP_MemoryOutDrv(20);
	_TP_MemoryOutDrv(tgid);
	_TP_MemoryOutDrv((tgid >> 8));
	_TP_MemoryOutDrv((tgid >> 16));
	_TP_MemoryOutDrv((tgid >> 24));
	for (i = 0; i < 16; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	STOP_CON;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusKernelInfo);
/********************************************************************************/

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int dt_proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct DT_DATA dt_data;

	switch (buf[0]) {
	case '0':
		start_trace = 0;
		printk( KERN_INFO "[DT] Trace Stop.\n" );
		break;
	case '1':
		start_trace = 1;
		printk( KERN_INFO "[DT] Trace Start.\n" );
		break;
	case TP_BusOut:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_BusOut(dt_data.addr, dt_data.dat);
		break;
	case TP_MemoryOutput:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_MemoryOutput(dt_data.addr, dt_data.dat, dt_data.value, dt_data.size);
		break;
	case TP_BusOutFast:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_BusOutFast(dt_data.addr, dt_data.dat);
		break;
	case TP_MemoryOutputFast:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_MemoryOutputFast(dt_data.addr, dt_data.dat, dt_data.value, dt_data.size);
		break;
	case TP_EventTrigger:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_EventTrigger(dt_data.addr, dt_data.dat, dt_data.size);
		break;
	case TP_EventTriggerFast:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_EventTriggerFast(dt_data.addr, dt_data.dat, dt_data.size);
		break;
	case TP_EventTrigger32:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_EventTrigger32(dt_data.addr, dt_data.dat, dt_data.size);
		break;
	case TP_EventTrigger32Fast:
		if (copy_from_user(&dt_data, buf, len)) {
			return len;
		}
		_TP_EventTrigger32Fast(dt_data.addr, dt_data.dat, dt_data.size);
		break;
	default:
		printk( KERN_INFO "[DT] Trace Command is 0-9.\n" );
		break;
	}

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_read(struct file *filp, char __user *buf, size_t count, loff_t *data)
#else
static int dt_proc_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
#endif
{
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	*eof = 1;
#endif
	return count;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static const struct file_operations dt_proc_fops = {
	.write = dt_proc_write,
	.read = dt_proc_read,
};
#endif

static int __init dt_proc_init(void)
{
	struct proc_dir_entry* entry;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry = proc_create(PROC_NAME, 666, NULL, &dt_proc_fops);
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#else
	entry = create_proc_entry(PROC_NAME, 666, NULL);
	entry->read_proc = dt_proc_read;
	entry->write_proc = dt_proc_write;
#endif
	_TP_BusPortInit();
	printk(KERN_INFO "dtproc load OK\n");
	if(init == 1){
		start_trace = 1;
	}
	return 0;
}

static void __exit dt_proc_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	if (init != 0) {
		start_trace = 0;
		init = 0;
		iounmap(iop);
	}
	printk(KERN_INFO "dtproc unload\n");
}

module_init(dt_proc_init);
module_exit(dt_proc_exit);

MODULE_DESCRIPTION("dtproc");
MODULE_LICENSE(LICENSE);

/********************************************************************************/

